use crate::constants::*;
use crate::errors::*;
use std::borrow::Cow;
use tokio::io::{AsyncRead, AsyncReadExt, AsyncWrite, AsyncWriteExt};

#[derive(Debug)]
pub(crate) enum Command<'a> {
	DataFrame(Cow<'a, [u8]>),
	TxDelay(u8),
	/*
	Persistence(u8),
	SlotTime(u8),
	TxTail(u8),
	FullDuplex(u8),
	*/
	Return,
}

impl<'a> Command<'a> {
	pub fn borrowed_frame(data: &'a [u8]) -> Self {
		Self::DataFrame(Cow::Borrowed(data))
	}

	pub(crate) async fn write<W: AsyncWrite + Unpin>(
		&self,
		port: u8,
		writer: &mut W,
	) -> tokio::io::Result<usize> {
		assert!(port < 16);
		let port = port << 4;

		let mut count = 0;
		writer.write_all(&[FEND]).await?;
		count += 1;
		match self {
			Self::DataFrame(data) => {
				writer.write_all(&[DATA_FRAME | port]).await?;
				count += 1;
				for &b in data.as_ref() {
					count += match b {
						FEND => {
							writer.write_all(&[FESC, TFEND]).await?;
							2
						}
						FESC => {
							writer.write_all(&[FESC, TFESC]).await?;
							2
						}
						_ => {
							writer.write_all(&[b]).await?;
							1
						}
					}
				}
			}

			Self::TxDelay(delay) => {
				writer.write_all(&[TX_DELAY | port, *delay]).await?;
				count += 2;
			}

			Self::Return => {
				writer.write_all(&[RETURN]).await?;
				count += 1;
			}
		}

		writer.write_all(&[FEND]).await?;
		count += 1;

		Ok(count)
	}

	pub async fn read<R: AsyncRead + Unpin>(reader: &mut R) -> Result<(u8, Vec<u8>), ReadError> {
		// BUG - Anything received on port 12 will cause issues
		// Because data frames will be marked as 0xC0 - the same as FEND
		// I'm not sure how many people actually use 12 ports, so this is
		// probably a non-issue.

		let mut buf = [0];
		reader.read_exact(&mut buf).await?;

		if buf != [0xC0] {
			// First byte must be FEND
			return Err(ReadError::MalformedData);
		}

		// Additional 0xC0 frames should be dropped
		while buf == [0xC0] {
			reader.read_exact(&mut buf).await?;
		}

		let port = (buf[0] & 0xF0) >> 4;
		let code = buf[0] & 0x0F;

		if code != 0x00 {
			// The TNC can only send us data frames
			return Err(ReadError::MalformedData);
		}

		let mut data = vec![];
		reader.read_exact(&mut buf).await?;
		while buf[0] != 0xC0 {
			let b = buf[0];
			if b == FESC {
				reader.read_exact(&mut buf).await?;
				match buf[0] {
					TFEND => {
						data.push(FEND);
					}
					TFESC => {
						data.push(FESC);
					}
					_ => return Err(ReadError::MalformedData),
				}
			} else {
				data.push(b);
			}

			reader.read_exact(&mut buf).await?;
		}

		Ok((port, data))
	}
}
