use crate::command::Command;
use crate::errors::*;
use tokio::io::{AsyncRead, AsyncWrite, AsyncWriteExt, BufReader, BufWriter};
use tokio::net::{
	tcp::{OwnedReadHalf, OwnedWriteHalf},
	TcpStream, ToSocketAddrs,
};

/// Handles reading from the Tnc
pub struct TncRead<R: AsyncRead + Unpin> {
	reader: BufReader<R>,
}

impl<R: AsyncRead + Unpin> TncRead<R> {
	/// Construct a new TncRead
	pub fn new(reader: R) -> Self {
		Self {
			reader: BufReader::new(reader),
		}
	}

	/// Reads a frame from the TNC. Returns a tuple containing the port number
	/// as well as the data contained in the frame.
	pub async fn read_frame(&mut self) -> Result<(u8, Vec<u8>), ReadError> {
		Command::read(&mut self.reader).await
	}
}

/// Handles writing to the Tnc
pub struct TncWrite<W: AsyncWrite + Unpin> {
	writer: BufWriter<W>,
	port: u8,
}

impl<W: AsyncWrite + Unpin> TncWrite<W> {
	/// Construct a new TncWrite
	pub fn new(writer: W) -> Self {
		Self {
			writer: BufWriter::new(writer),
			port: 0,
		}
	}

	/// Sets the port to use for transmissions.
	/// # Arguments
	///
	/// * port - The port number. Valid values are 0 to 15.
	///
	/// Panics if port number is invalid.
	pub fn set_port(&mut self, port: u8) {
		if port > 15 {
			panic!("Port number is invalid");
		}

		self.port = port;
	}

	/// Send an array of bytes as a frame. Note that the bytes are not split up if there are a lot of them.
	pub async fn send_frame(&mut self, data: &[u8]) -> std::io::Result<usize> {
		let command = Command::borrowed_frame(data);
		command.write(self.port, &mut self.writer).await
	}

	/// Sets the amount of time between keying the transmitter and beginning to send data
	/// # Arguments
	///
	/// * delay - The amount of time to wait, in 10 ms units
	pub async fn set_tx_delay(&mut self, delay: u8) -> std::io::Result<usize> {
		let command = Command::TxDelay(delay);
		command.write(self.port, &mut self.writer).await
	}

	/// Exits KISS mode
	///
	/// NOTE: After exiting KISS mode the reader could do anything.
	/// This is because the TNC could return non-KISS data which
	/// may or may not be interpreted as valid KISS commands
	pub async fn exit(mut self) -> tokio::io::Result<usize> {
		let command = Command::Return;
		let c = command.write(self.port, &mut self.writer).await?;
		self.flush().await?;
		Ok(c)
	}

	/// Flushes the transmission buffer.
	pub async fn flush(&mut self) -> tokio::io::Result<()> {
		self.writer.flush().await
	}
}

pub struct Tnc<R: AsyncRead + Unpin, W: AsyncWrite + Unpin> {
	read: TncRead<R>,
	write: TncWrite<W>,
}

impl Tnc<OwnedReadHalf, OwnedWriteHalf> {
	/// Connects to a KISS TNC over TCP.
	/// # Arguments
	///
	/// * addr - The address to connect to.
	pub async fn connect_tcp<A: ToSocketAddrs>(addr: A) -> std::io::Result<Self> {
		let stream = TcpStream::connect(addr).await?;

		let (r, w) = stream.into_split();

		Ok(Self::new(r, w))
	}
}

impl<R: AsyncRead + Unpin, W: AsyncWrite + Unpin> Tnc<R, W> {
	/// Construct a Tnc from a reader and a writer
	pub fn new(reader: R, writer: W) -> Self {
		Self {
			read: TncRead::new(reader),
			write: TncWrite::new(writer),
		}
	}

	/// Split Tnc into a TncRead and TncWrite
	pub fn split(self) -> (TncRead<R>, TncWrite<W>) {
		(self.read, self.write)
	}

	/// Sets the port to use for transmissions.
	/// # Arguments
	///
	/// * port - The port number. Valid values are 0 to 15.
	///
	/// Panics if port number is invalid.
	pub fn set_port(&mut self, port: u8) {
		self.write.set_port(port);
	}

	/// Send an array of bytes as a frame. Note that the bytes are not split up if there are a lot of them.
	pub async fn send_frame(&mut self, data: &[u8]) -> std::io::Result<usize> {
		self.write.send_frame(data).await
	}

	/// Sets the amount of time between keying the transmitter and beginning to send data
	/// # Arguments
	///
	/// * delay - The amount of time to wait, in 10 ms units
	pub async fn set_tx_delay(&mut self, delay: u8) -> std::io::Result<usize> {
		self.write.set_tx_delay(delay).await
	}

	/// Exits KISS mode
	pub async fn exit(self) -> tokio::io::Result<usize> {
		self.write.exit().await
	}

	/// Flushes the transmission buffer.
	pub async fn flush(&mut self) -> tokio::io::Result<()> {
		self.write.flush().await
	}

	/// Reads a frame from the TNC. Returns a tuple containing the port number
	/// as well as the data contained in the frame.
	pub async fn read_frame(&mut self) -> Result<(u8, Vec<u8>), ReadError> {
		self.read.read_frame().await
	}
}

#[cfg(test)]
mod tests {
	use super::*;

	#[tokio::test]
	async fn test_send_frame_works_corrently() {
		let mut tnc = Tnc::new(&[][..], vec![]);

		assert_eq!(7, tnc.send_frame("TEST".as_bytes()).await.unwrap());
		tnc.flush().await.unwrap();
		assert_eq!(
			vec![0xC0, 0x00, 0x54, 0x45, 0x53, 0x54, 0xC0],
			tnc.write.writer.into_inner()
		);

		let mut tnc = Tnc::new(&[][..], vec![]);

		tnc.set_port(5);
		assert_eq!(8, tnc.send_frame("Hello".as_bytes()).await.unwrap());
		tnc.flush().await.unwrap();
		assert_eq!(
			vec![0xC0, 0x50, 0x48, 0x65, 0x6C, 0x6C, 0x6F, 0xC0],
			tnc.write.writer.into_inner()
		);

		let mut tnc = Tnc::new(&[][..], vec![]);

		assert_eq!(7, tnc.send_frame(&[0xC0, 0xDB]).await.unwrap());
		tnc.flush().await.unwrap();
		assert_eq!(
			vec![0xC0, 0x00, 0xDB, 0xDC, 0xDB, 0xDD, 0xC0],
			tnc.write.writer.into_inner()
		);
	}

	#[tokio::test]
	async fn test_set_tx_delay_works_correctly() {
		let mut tnc = Tnc::new(&[][..], vec![]);
		tnc.set_port(14);

		// Delay of 100 * 10 ms = 1 second
		assert_eq!(4, tnc.set_tx_delay(100).await.unwrap());
		tnc.flush().await.unwrap();
		assert_eq!(vec![0xC0, 0xE1, 100, 0xC0], tnc.write.writer.into_inner());
	}

	#[tokio::test]
	async fn test_read_frame_works_correctly() {
		let mut tnc = Tnc::new(&[0xC0, 0x00, 0x54, 0x45, 0x53, 0x54, 0xC0][..], vec![]);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(0, port);
		assert_eq!("TEST".as_bytes(), data);
	}

	#[tokio::test]
	async fn test_read_frame_sets_port_correctly() {
		let mut tnc = Tnc::new(
			&[0xC0, 0x50, 0x48, 0x65, 0x6C, 0x6C, 0x6F, 0xC0][..],
			vec![],
		);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(5, port);
		assert_eq!("Hello".as_bytes(), data);
	}

	#[tokio::test]
	async fn test_read_frame_multiple_frames() {
		let mut tnc = Tnc::new(
			&[
				0xC0, 0x50, 0x48, 0x65, 0x6C, 0x6C, 0x6F, 0xC0, 0xC0, 0x00, 0x54, 0x45, 0x53, 0x54,
				0xC0,
			][..],
			vec![],
		);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(5, port);
		assert_eq!("Hello".as_bytes(), data);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(0, port);
		assert_eq!("TEST".as_bytes(), data);
	}

	#[tokio::test]
	async fn test_read_frame_escape_works_correctly() {
		let mut tnc = Tnc::new(&[0xC0, 0x00, 0xDB, 0xDC, 0xDB, 0xDD, 0xC0][..], vec![]);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(0, port);
		assert_eq!(vec![0xC0, 0xDB], data);
	}

	#[tokio::test]
	async fn test_read_frame_ignores_additional_c0s() {
		let mut tnc = Tnc::new(
			&[
				0xC0, 0xC0, 0xC0, 0xC0, 0xC0, 0x00, 0xDB, 0xDC, 0xDB, 0xDD, 0xC0,
			][..],
			vec![],
		);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(0, port);
		assert_eq!(vec![0xC0, 0xDB], data);
	}

	#[tokio::test]
	async fn test_read_frame_requires_c0() {
		let mut tnc = Tnc::new(&[0x00, 0xDB, 0xDC, 0xDB, 0xDD, 0xC0][..], vec![]);

		assert!(matches!(
			tnc.read_frame().await.unwrap_err(),
			ReadError::MalformedData
		));
	}

	#[tokio::test]
	async fn test_bytes_after_c0_are_ignored() {
		let mut tnc = Tnc::new(
			&[
				0xC0, 0x00, 0xDB, 0xDC, 0xDB, 0xDD, 0xC0, 0xA, 0xB, 0xC, 0xAB, 0xC0, 0x12, 0x53,
			][..],
			vec![],
		);

		let (port, data) = tnc.read_frame().await.unwrap();
		assert_eq!(0, port);
		assert_eq!(vec![0xC0, 0xDB], data);
	}
}
